from enum import Enum
from typing import List


class StrEnum(str, Enum):
    def __str__(self) -> str:
        """Used when dumping enum fields in a schema."""
        ret: str = self.value
        return ret

    @classmethod
    def list(cls) -> List[str]:
        return [c.value for c in cls]



from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple
from urllib.parse import urlencode
from uuid import UUID


def convert_to_serializable(value: Any) -> Any:
    """Convert values to a serializable format."""
    if isinstance(value, datetime):
        return value.isoformat()
    if isinstance(value, UUID):
        return str(value)
    if isinstance(value, (List, Tuple)):
        return [convert_to_serializable(v) for v in value]
    if isinstance(value, Dict):
        return {k: convert_to_serializable(v) for k, v in value.items()}
    return value


def build_query_params(
    filters: Optional[Dict[str, Any]] = None,
    pagination_params: Optional[Dict[str, int]] = None,
) -> str:
    """Convert filters to a URL-encoded query parameter string.

    Encode query parameters using doseq=True to handle sequences
    correctly, it ensures that each element of a sequence
    is treated as a separate query parameter.
    """
    serialised_params = {}
    if filters:
        serialised_params.update(convert_to_serializable(filters))
    if pagination_params:
        serialised_params.update(pagination_params)
    # To avoid returning a single interrogation point without query parameters
    return "?" + urlencode(serialised_params, doseq=True) if serialised_params else ""



from __future__ import annotations

from dataclasses import asdict, dataclass, fields
from typing import Any, Dict, List, Optional

from pasqal_cloud.device.configuration.result_type import ResultType


class InvalidConfiguration(Exception):
    pass


INVALID_KEY_ERROR_MSG = (
    "Invalid key {} in Configuration.extra_config. "
    + "Attempted to override a default field."
)
INVALID_RESULT_TYPES = (
    "Invalid set result types {} in Configuration. This device allows for {}."
)


@dataclass
class BaseConfig:
    """Base class for all device configuration classes.

    This class provides a common interface for all device configuration classes.
    It also provides a mechanism to add extra configuration parameters
    that are not part of the default configuration.

    The extra configuration parameters are stored in the `extra_config` field.
    This field is a dictionary that can be used to store any extra configuration
    parameters. The `extra_config` field is not part of the configuration
    dictionary returned by the `to_dict` method. Instead, the extra configuration
    parameters are added to the configuration dictionary.
    """

    strict_validation: bool = False
    """Whether to validate an emulator job as though it were for FRESNEL."""
    extra_config: Optional[Dict[str, Any]] = None
    result_types: Optional[List[ResultType]] = None

    def to_dict(self) -> dict[str, Any]:
        """Converts the configuration to a dictionary.

        The extra configuration parameters are added to the configuration
        dictionary.

        Returns:
            dict: Configuration dictionary.
        """
        self._validate()
        return self._unnest_extra_config(asdict(self))

    @classmethod
    def from_dict(cls, conf: dict[str, Any]) -> BaseConfig:
        """Creates a configuration object from a dictionary.

        The extra configuration parameters are stored in the `extra_config` field.

        Args:
            conf (dict): Configuration dictionary.

        Returns:
            BaseConfig: Configuration object.
        """
        base_conf = {}
        for field in fields(cls):
            if field.name != "extra_config" and field.name in conf:
                base_conf[field.name] = conf.pop(field.name)

        # ensure that no extra config is passed as None
        if not conf:
            conf = None  # type: ignore[assignment]
        return cls(**base_conf, extra_config=conf)

    @staticmethod
    def _unnest_extra_config(conf_dict: dict[str, Any]) -> dict[str, Any]:
        res = {k: v for (k, v) in conf_dict.items() if k != "extra_config"}
        if conf_dict.get("extra_config"):
            res.update(conf_dict["extra_config"])
        return res

    @property
    def allowed_result_types(self) -> List[ResultType]:
        return [ResultType.COUNTER]

    def _validate(self) -> None:
        if self.extra_config:
            for k in self.extra_config:
                if k in self.__dataclass_fields__:
                    raise InvalidConfiguration(INVALID_KEY_ERROR_MSG.format(k))
        if self.result_types and not set(self.result_types) <= set(
            self.allowed_result_types
        ):
            raise InvalidConfiguration(
                INVALID_RESULT_TYPES.format(
                    self.result_types, self.allowed_result_types
                )
            )



import json
import os
from typing import Any, Dict, Generator, Optional
from unittest.mock import patch

import pytest
import requests_mock
from requests import HTTPError, Request

from pasqal_cloud import Batch, Client, Job, Workload
from pasqal_cloud.endpoints import Endpoints
from tests.test_doubles.authentication import FakeAuth0AuthenticationSuccess

TEST_API_FIXTURES_PATH = "tests/fixtures/api"
RESULT_LINK_ENDPOINT = "http://result-link/"
JOB_RESULT_LINK_ENDPOINT = "http://jobs-result-link/"
JSON_FILE = "_.{}.json"


@pytest.fixture()
def result_link_endpoint() -> str:
    return RESULT_LINK_ENDPOINT


def mock_core_response(request):
    version = request.url.split("api/")[1].split("/")[0]
    path = request.url.split(f"{version}/")[1].split("?")[0]
    data = None
    if request.method == "POST" and request.body:
        data = request.json()

    json_path = os.path.join(
        TEST_API_FIXTURES_PATH, version, path, JSON_FILE.format(request.method)
    )
    with open(json_path) as json_file:
        result = json.load(json_file)
        if path == "batches" and data:
            if data.get("emulator"):
                result["data"]["device_type"] = data["emulator"]
            else:
                result["data"]["device_type"] = "FRESNEL"
                result["data"]["configuration"] = None
        return result


def mock_s3_presigned_url_response(request: Request) -> Dict[str, Any]:
    """
    Mock the response of a presigned URL and return expected results

    Args:
        request (Request): The incoming HTTP request

    Returns:
        Dict[str, Any]: The parsed JSON content from the results file
    """
    resource_id = request.url.split("/")[-1]
    results_path = os.path.join(
        TEST_API_FIXTURES_PATH, "v1", f"jobs/{resource_id}/results_link/result.json"
    )
    with open(results_path) as json_file:
        return json.load(json_file)


def mock_result_link_response() -> Dict[str, str]:
    """This mocks the response from the s3 result link."""
    return {"some": "result"}


def mock_response(request, _) -> Dict[str, Any]:
    """This acts as a Router to Mock the requests we make with custom behaviors.

    A linter might suggest 'context' is unused, but is required for some
    tests to execute.
    """
    if request.url.startswith(Endpoints.core):
        return mock_core_response(request)
    if request.url.startswith(JOB_RESULT_LINK_ENDPOINT):
        return mock_s3_presigned_url_response(request)
    if request.url.startswith(RESULT_LINK_ENDPOINT):
        return mock_result_link_response()
    return None


@pytest.fixture(scope="session")
@requests_mock.Mocker(kw="mock")
def request_mock(mock=None) -> Optional[Any]:
    # Configure requests to use the local JSON files a response
    mock.register_uri(
        requests_mock.ANY,
        requests_mock.ANY,
        status_code=200,
        json=mock_response,
    )
    return mock


@pytest.fixture(scope="session")
@requests_mock.Mocker(kw="mock")
def request_mock_exception(mock=None) -> Optional[Any]:
    mock.register_uri(
        requests_mock.ANY,
        requests_mock.ANY,
        exc=HTTPError,
    )
    return mock


@pytest.fixture(scope="session")
def batch_creation_error_data() -> Dict[str, Any]:
    return {
        "status": "fail",
        "message": "Unprocessable Entity",
        "code": "422",
        "data": {
            "description": [
                {
                    "loc": ["body", "jobs"],
                    "msg": "The batch sequence is not parametrized and"
                    + "expects no variables.",
                    "type": "value_error",
                }
            ]
        },
    }


@pytest.fixture()
def mock_request(request_mock) -> Generator[Any, Any, None]:
    request_mock.start()
    yield request_mock
    request_mock.stop()


@pytest.fixture()
def mock_request_exception(request_mock_exception) -> Generator[Any, Any, None]:
    request_mock_exception.start()
    yield request_mock_exception
    request_mock_exception.stop()


@pytest.fixture()
@patch("pasqal_cloud.client.Auth0TokenProvider", FakeAuth0AuthenticationSuccess)
def pasqal_client_mock():
    return Client(
        project_id="00000000-0000-0000-0000-000000000002",
        username="00000000-0000-0000-0000-000000000001",
        password="password",
    )


@pytest.fixture()
def workload(pasqal_client_mock):
    workload_data = {
        "id": "00000000-0000-0000-0000-000000000001",
        "project_id": "00000000-0000-0000-0000-000000000001",
        "status": "PENDING",
        "_client": pasqal_client_mock,
        "created_at": "2022-12-31T23:59:59.999Z",
        "updated_at": "2023-01-01T00:00:00.000Z",
        "errors": [],
        "backend": "backend_test",
        "workload_type": "workload_type_test",
        "config": {"test1": "test1", "test2": 2},
    }
    return Workload(**workload_data)


@pytest.fixture()
def batch_data_fixture() -> Dict[str, Any]:
    return {
        "complete": False,
        "open": True,
        "created_at": "2022-12-31T23:59:59.999Z",
        "updated_at": "2023-01-01T00:00:00.000Z",
        "device_type": "qpu",
        "project_id": "00000000-0000-0000-0000-000000000002",
        "id": "00000000-0000-0000-0000-000000000001",
        "user_id": "EQZj1ZQE",
        "status": "PENDING",
        "webhook": "https://example.com/webhook",
        "sequence_builder": "pulser",
        "start_datetime": "2023-01-01T00:00:00.000Z",
        "end_datetime": None,
        "device_status": "available",
    }


@pytest.fixture()
def batch(batch_data_fixture: Dict[str, Any], pasqal_client_mock) -> Batch:
    return Batch(**batch_data_fixture, _client=pasqal_client_mock)


@pytest.fixture()
def job(pasqal_client_mock):
    job_data = {
        "runs": 50,
        "batch_id": "00000000-0000-0000-0000-000000000001",
        "id": "00000000-0000-0000-0000-000000022010",
        "project_id": "00000000-0000-0000-0000-000000000001",
        "status": "PENDING",
        "_client": pasqal_client_mock,
        "created_at": "2022-12-31T23:59:59.999Z",
        "updated_at": "2023-01-01T00:00:00.000Z",
        "errors": [],
        "variables": {"param1": 1, "param2": 2, "param3": 3},
    }
    return Job(**job_data)



# ruff: noqa: E402 F403

import warnings


def pasqal_sdk_not_updated():
    warnings.warn(
        "pasqal-sdk package is deprecated, please use pasqal_cloud"
        "instead: `pip install pasqal-cloud`",
        DeprecationWarning,
    )


pasqal_sdk_not_updated()
from pasqal_cloud import *



import time
from typing import Any, Dict, List, Optional, Type, Union
from warnings import warn

from pydantic import BaseModel, ConfigDict, field_validator, PrivateAttr, ValidationInfo
from requests import HTTPError

from pasqal_cloud.client import Client
from pasqal_cloud.device import EmulatorType
from pasqal_cloud.device.configuration import BaseConfig, EmuFreeConfig, EmuTNConfig
from pasqal_cloud.errors import (
    BatchCancellingError,
    BatchClosingError,
    BatchFetchingError,
    JobCreationError,
    JobRetryError,
)
from pasqal_cloud.job import CreateJob, Job

RESULT_POLLING_INTERVAL = 2  # seconds


class Batch(BaseModel):
    """Class to load batch data return by the API.

    A batch groups up several jobs with the same sequence. When a batch is assigned to
    a QPU, all its jobs are run sequentially and no other batch can be assigned to the
    device until all its jobs are done and declared complete.

    Attributes:
        - open: Whether the batch accepts more jobs or not.
        - complete: Opposite of open, deprecated.
        - created_at: Timestamp of the creation of the batch.
        - updated_at: Timestamp of the last update of the batch.
        - device_type: Type of device to run the batch on.
        - project_id: ID of the owner project of the batch.
        - id: Unique identifier for the batch.
        - user_id: Unique identifier of the user that created the batch.
        - status: Status of the batch. Possible values are:
            PENDING, RUNNING, DONE, CANCELED, TIMED_OUT, ERROR, PAUSED.
        - webhook: Webhook where the job results are automatically sent to.
        - _client: A Client instance to connect to PCS.
        - sequence_builder: Pulser sequence of the batch.
        - start_datetime: Timestamp of the time the batch was sent to the QPU.
        - end_datetime: Timestamp of when the batch process was finished.
        - device_status: Status of the device where the batch is running.
        - jobs (deprecated): Dictionary of all the jobs added to the batch.
        - ordered_jobs: List of all the jobs added to the batch,
            ordered by creation time.
        - jobs_count: Number of jobs added to the batch.
        - jobs_count_per_status: Number of jobs per status.
        - configuration: Further configuration for certain emulators.
        - group_id (deprecated): Use project_id instead.
    """

    open: bool
    complete: bool
    created_at: str
    updated_at: str
    device_type: str
    project_id: str
    id: str
    user_id: str
    status: str
    _client: Client = PrivateAttr(default=None)
    sequence_builder: str
    _ordered_jobs: Optional[List[Job]] = PrivateAttr(default=None)
    jobs_count: int = 0
    jobs_count_per_status: Dict[str, int] = {}
    webhook: Optional[str] = None
    start_datetime: Optional[str] = None
    end_datetime: Optional[str] = None
    device_status: Optional[str] = None
    parent_id: Optional[str] = None
    configuration: Union[BaseConfig, Dict[str, Any], None] = None

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    def __init__(self, **data: Any) -> None:
        """
        Workaround to make the private attribute '_client' working
        like we need with Pydantic V2, more information on:
        https://docs.pydantic.dev/latest/concepts/models/#private-model-attributes
        """
        super().__init__(**data)
        self._client = data["_client"]
        if data.get("jobs"):
            self._ordered_jobs = [
                Job(**raw_job, _client=self._client) for raw_job in data["jobs"]
            ]

    @property
    def ordered_jobs(self) -> List[Job]:
        if self._ordered_jobs is None:
            raw_jobs = self._client.get_batch_jobs(batch_id=self.id)
            self._ordered_jobs = [
                Job(**raw_job, _client=self._client) for raw_job in raw_jobs
            ]
        return self._ordered_jobs

    @ordered_jobs.setter
    def ordered_jobs(self, value: Any) -> None:
        self._ordered_jobs = value

    # Ticket (#704), to be removed or updated
    @property
    def jobs(self) -> Dict[str, Job]:
        """Once the 'ordered_jobs' is built, we need to keep the 'jobs' attribute
        for backward compatibility with the code written by the users of the sdk
        """
        warn(
            "'jobs' attribute is deprecated, use 'ordered_jobs' instead",
            DeprecationWarning,
            stacklevel=1,
        )
        return {job.id: job for job in self.ordered_jobs}

    @jobs.setter
    def jobs(self, _: Any) -> None:
        # Override the jobs setter to be a no-op.
        # `jobs` is a read-only attribute which is derived from the `ordered_jobs` key.
        pass

    @field_validator("configuration", mode="before")
    def _load_configuration(
        cls,
        configuration: Union[Dict[str, Any], BaseConfig, None],
        info: ValidationInfo,
    ) -> Optional[BaseConfig]:
        if not isinstance(configuration, dict):
            return configuration
        conf_class: Type[BaseConfig] = BaseConfig
        if info.data["device_type"] == EmulatorType.EMU_TN.value:
            conf_class = EmuTNConfig
        elif info.data["device_type"] == EmulatorType.EMU_FREE.value:
            conf_class = EmuFreeConfig
        return conf_class.from_dict(configuration)

    def add_jobs(
        self,
        jobs: List[CreateJob],
        wait: bool = False,
    ) -> None:
        """Add some jobs to batch for execution on PCS and returns the updated batch.

        The batch must be `open` otherwise the API will return an error.
        The new jobs are appended to the `ordered_jobs` list attribute.

        Args:
            jobs: List of jobs to be added to the batch.
            wait: If True, blocks until all jobs in the batch are done.

        """
        try:
            batch_rsp = self._client.add_jobs(self.id, jobs)
        except HTTPError as e:
            raise JobCreationError(e) from e
        self._update_from_api_response(batch_rsp)

        if wait:
            while any(
                job.status in {"PENDING", "RUNNING"} for job in self.ordered_jobs
            ):
                time.sleep(RESULT_POLLING_INTERVAL)
                self.refresh()

    def retry(self, job: Job, wait: bool = False) -> None:
        """
        Retry a job in the same batch.
        The batch should not be 'closed'.
        The new job is appended to the `ordered_jobs` list attribute.

        Args:
            job: The job to retry
            wait: Whether to wait for job completion

        Raises:
            JobRetryError if there was an error adding the job to the batch.
        """
        retried_job = CreateJob(runs=job.runs, variables=job.variables)
        try:
            self.add_jobs([retried_job], wait=wait)
        except JobCreationError as e:
            raise JobRetryError from e

    def declare_complete(self, wait: bool = False, fetch_results: bool = False) -> None:
        """
        Deprecated, use close instead.
        """
        warn(
            "This method is deprecated, use close instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.close(wait=wait, fetch_results=fetch_results)

    def close(self, wait: bool = False, fetch_results: bool = False) -> None:
        """Declare to PCS that the batch is closed and returns an updated
        batch instance.

        Args:
            wait: Whether to wait for the batch to be done and fetch results.
            fetch_results (deprecated): Whether to wait for the batch \
                to be done and fetch results.

        A batch that is closed awaits no extra jobs. All jobs previously added
        will be executed before the batch is terminated. When all its jobs are done,
        the closed batch is unassigned to its running device.
        """
        try:
            batch_rsp = self._client.close_batch(self.id)
        except HTTPError as e:
            raise BatchClosingError(e) from e
        self._update_from_api_response(batch_rsp)
        if wait or fetch_results:
            while any(
                job.status in {"PENDING", "RUNNING"} for job in self.ordered_jobs
            ):
                time.sleep(RESULT_POLLING_INTERVAL)
                self.refresh()

    def cancel(self) -> None:
        """Cancel the current batch on the PCS."""
        try:
            batch_rsp = self._client.cancel_batch(self.id)
        except HTTPError as e:
            raise BatchCancellingError(e) from e
        self._update_from_api_response(batch_rsp)

    def refresh(self) -> None:
        """Fetch the batch from the API and update it in place."""
        try:
            batch_rsp = self._client.get_batch(self.id)
        except HTTPError as e:
            raise BatchFetchingError(e) from e
        self._update_from_api_response(batch_rsp)

    def _update_from_api_response(self, data: Dict[str, Any]) -> None:
        """Update the instance in place with the response body of the batch API"""
        updated_batch = Batch(**data, _client=self._client)
        # Private fields are not exposed by the model_fields method, so we need to
        # explicitly add _ordered_jobs to ensure it is copied from the updated_batch
        # model.
        batch_model_fields = [*list(updated_batch.model_fields), "_ordered_jobs"]
        for field in batch_model_fields:
            value = getattr(updated_batch, field)
            setattr(self, field, value)



# Copyright 2020 Pasqal Cloud Services development team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from getpass import getpass
from typing import Any, Dict, List, Mapping, Optional, Sequence, Union
from uuid import UUID

import requests
from requests.auth import AuthBase

from pasqal_cloud._version import __version__ as sdk_version 
from pasqal_cloud.authentication import (
    Auth0TokenProvider,
    HTTPBearerAuthenticator,
    TokenProvider,
)
from pasqal_cloud.endpoints import Auth0Conf, Endpoints
from pasqal_cloud.utils.filters import (
    CancelJobFilters,
    JobFilters,
    PaginationParams,
    RebatchFilters,
)
from pasqal_cloud.utils.jsend import JobResult, JSendPayload
from pasqal_cloud.utils.retry import retry_http_error


TIMEOUT = 30  # client http requests timeout after 30s


class EmptyFilter:
    pass


class Client:
    authenticator: AuthBase

    def __init__(
        self,
        project_id: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token_provider: Optional[TokenProvider] = None,
        endpoints: Optional[Endpoints] = None,
        auth0: Optional[Auth0Conf] = None,
    ):
        if not username and not token_provider:
            raise ValueError(
                "At least a username or TokenProvider object should be provided."
            )
        if token_provider is not None:
            self._check_token_provider(token_provider)

        self.endpoints = self._make_endpoints(endpoints)

        if username:
            auth0 = self._make_auth0(auth0)
            token_provider = self._credential_login(username, password, auth0)

        self.authenticator = HTTPBearerAuthenticator(token_provider)
        self.project_id = project_id
        self.user_agent = f"PasqalCloudSDK/{sdk_version}"

    @staticmethod
    def _make_endpoints(endpoints: Optional[Endpoints]) -> Endpoints:
        if endpoints is None:
            return Endpoints()

        if not isinstance(endpoints, Endpoints):
            raise TypeError(f"endpoints must be a {Endpoints.__name__} instance")

        return endpoints

    @staticmethod
    def _make_auth0(auth0: Optional[Auth0Conf]) -> Auth0Conf:
        if auth0 is None:
            return Auth0Conf()

        if not isinstance(auth0, Auth0Conf):
            raise TypeError(f"auth0 parameter must be a {Auth0Conf.__name__} instance")

        return auth0

    @staticmethod
    def _check_token_provider(token_provider: TokenProvider) -> None:
        if not isinstance(token_provider, TokenProvider):
            raise TypeError("token_provider must be an instance of TokenProvider.")

    def _credential_login(
        self, username: str, password: Optional[str], auth0: Auth0Conf
    ) -> TokenProvider:
        if not password:
            password = getpass("Enter your password:")
            #   We want to allow an empty string at first, but then
            #   it results in an error
            if not password:
                raise ValueError("The prompted password should not be empty")

        token_provider: TokenProvider = Auth0TokenProvider(username, password, auth0)
        return token_provider

    @retry_http_error(
        max_retries=5, retry_status_code={408, 425, 429, 500, 502, 503, 504}
    )
    def _authenticated_request(
        self,
        method: str,
        url: str,
        payload: Optional[Union[Mapping, Sequence[Mapping]]] = None,
        params: Optional[Mapping[str, Any]] = None,
    ) -> JSendPayload:
        resp = requests.request(
            method,
            url,
            json=payload,
            timeout=TIMEOUT,
            headers={
                "content-type": "application/json", 
                "User-Agent": self.user_agent,
            },
            auth=self.authenticator,
            params=params,
        )
        resp.raise_for_status()
        data: JSendPayload = resp.json()
        return data

    def _request_all_pages(
        self,
        method: str,
        url: str,
        payload: Optional[Union[Mapping, Sequence[Mapping]]] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> List[Dict[str, Any]]:
        """
        Request all pages of a paginated endpoint and return a list of all items from
        the requested pages

        Args:
            method: HTTP method
            url: requested endpoint url
            payload: query payload
            params: query params

        Returns:
            A list of items from the requested pages.
        """

        if not params:
            params = {}
        first_page_response = self._authenticated_request(
            method=method, url=url, payload=payload, params=params
        )
        all_items: List[Dict] = first_page_response["data"]
        pagination_data = first_page_response.get("pagination")

        if not pagination_data:
            return all_items

        total_nb_items = pagination_data["total"]
        end = pagination_data["end"]
        while end < total_nb_items:
            params["offset"] = end
            response: JSendPayload = self._authenticated_request(
                method=method, url=url, payload=payload, params=params
            )

            all_items.extend(response["data"])
            pagination_data = response.get("pagination")

            # Added to make sure the type is correctly set for mypy
            # Should not happen
            if not pagination_data:
                break

            total_nb_items = pagination_data["total"]
            end = pagination_data["end"]
        return all_items

    def send_batch(self, batch_data: Dict[str, Any]) -> Dict[str, Any]:
        batch_data.update({"project_id": self.project_id})
        response: Dict[str, Any] = self._authenticated_request(
            "POST",
            f"{self.endpoints.core}/api/v1/batches",
            batch_data,
        )["data"]
        return response

    def get_batch(self, batch_id: str) -> Dict[str, Any]:
        response: Dict[str, Any] = self._authenticated_request(
            "GET", f"{self.endpoints.core}/api/v2/batches/{batch_id}"
        )["data"]
        return response

    def get_batch_jobs(self, batch_id: str) -> list[Dict[str, Any]]:
        return self._request_all_pages(
            "GET",
            f"{self.endpoints.core}/api/v2/jobs",
            params={
                "batch_id": batch_id,
                "order_by": "ordered_id",
                "order_by_direction": "ASC",
            },
            # The order-by parameters are required to ensure that the jobs appear in
            # the same order as they do in the GET /batches/{id} response
        )

    @retry_http_error(max_retries=5, retry_exceptions=(requests.ConnectionError,))
    def _download_results(self, results_link: str) -> JobResult:
        response = requests.request("GET", results_link)
        response.raise_for_status()
        data = response.json()
        return JobResult(
            raw=data.pop("raw", None), counter=data.pop("counter", None), **data
        )

    def get_job_results(self, job_id: str) -> Optional[JobResult]:
        """
        Download job result from S3.

        This function request a presigned-url for a specific job_id from the core API.
        Once the presigned-url is obtained, it downloads the result from S3.
        """
        results_link = self._authenticated_request(
            "GET", f"{self.endpoints.core}/api/v1/jobs/{job_id}/results_link"
        )["data"]["results_link"]
        if results_link:
            return self._download_results(results_link)
        return None

    def close_batch(self, batch_id: str) -> Dict[str, Any]:
        response: Dict[str, Any] = self._authenticated_request(
            "PUT", f"{self.endpoints.core}/api/v1/batches/{batch_id}/complete"
        )["data"]
        return response

    def cancel_batch(self, batch_id: str) -> Dict[str, Any]:
        response: Dict[str, Any] = self._authenticated_request(
            "PUT", f"{self.endpoints.core}/api/v1/batches/{batch_id}/cancel"
        )["data"]
        return response

    def rebatch(
        self, batch_id: Union[UUID, str], filters: RebatchFilters
    ) -> Dict[str, Any]:
        response: Dict[str, Any] = self._authenticated_request(
            "POST",
            f"{self.endpoints.core}/api/v1/batches/{batch_id}/rebatch",
            params=filters.model_dump(exclude_unset=True),
        )["data"]
        return response

    def get_jobs(
        self, filters: JobFilters, pagination_params: PaginationParams
    ) -> JSendPayload:
        filters_params = filters.model_dump(exclude_unset=True)
        filters_params.update(pagination_params.model_dump())
        response: JSendPayload = self._authenticated_request(
            "GET",
            f"{self.endpoints.core}/api/v2/jobs",
            params=filters_params,
        )
        return response

    def add_jobs(
        self, batch_id: str, jobs_data: Sequence[Mapping[str, Any]]
    ) -> Dict[str, Any]:
        response: Dict[str, Any] = self._authenticated_request(
            "POST", f"{self.endpoints.core}/api/v1/batches/{batch_id}/jobs", jobs_data
        )["data"]
        return response

    def get_job(self, job_id: str) -> Dict[str, Any]:
        response: Dict[str, Any] = self._authenticated_request(
            "GET", f"{self.endpoints.core}/api/v2/jobs/{job_id}"
        )["data"]
        return response

    def cancel_job(self, job_id: str) -> Dict[str, Any]:
        response: Dict[str, Any] = self._authenticated_request(
            "PUT", f"{self.endpoints.core}/api/v1/jobs/{job_id}/cancel"
        )["data"]
        return response

    def cancel_jobs(
        self, batch_id: Union[UUID, str], filters: CancelJobFilters
    ) -> Dict[str, Any]:
        response: Dict[str, Any] = self._authenticated_request(
            "PUT",
            f"{self.endpoints.core}/api/v1/batches/{batch_id}/cancel/jobs",
            params=filters.model_dump(exclude_unset=True),
        )["data"]
        return response

    def send_workload(self, workload_data: Dict[str, Any]) -> Dict[str, Any]:
        workload_data.update({"project_id": self.project_id})
        response: Dict[str, Any] = self._authenticated_request(
            "POST",
            f"{self.endpoints.core}/api/v1/workloads",
            workload_data,
        )["data"]
        return response

    def get_workload(self, workload_id: str) -> Dict[str, Any]:
        response: Dict[str, Any] = self._authenticated_request(
            "GET", f"{self.endpoints.core}/api/v2/workloads/{workload_id}"
        )["data"]
        return response

    def cancel_workload(self, workload_id: str) -> Dict[str, Any]:
        response: Dict[str, Any] = self._authenticated_request(
            "PUT", f"{self.endpoints.core}/api/v1/workloads/{workload_id}/cancel"
        )["data"]
        return response

    def get_device_specs_dict(self) -> Dict[str, str]:
        response: Dict[str, str] = self._authenticated_request(
            "GET", f"{self.endpoints.core}/api/v1/devices/specs"
        )["data"]
        return response



from unittest.mock import patch

import pytest

from pasqal_cloud import SDK
from tests.test_doubles.authentication import FakeAuth0AuthenticationSuccess

"""
Ticket (#622), Python file to be entirely removed
"""


@patch("pasqal_cloud.client.Auth0TokenProvider", FakeAuth0AuthenticationSuccess)
class TestInitSDKWithGroupId:
    project_id = "random_project_id"
    group_id = "random_group_id"
    username = "random_username"
    password = "random_password"

    @pytest.mark.filterwarnings(
        "ignore:The parameter 'group_id' is deprecated, from now on use 'project_id'"
        " instead"
    )
    def test_client_group_parameter(self):
        sdk = SDK(
            group_id=self.group_id, username=self.username, password=self.password
        )
        assert sdk._client.project_id == self.group_id

    def test_client_project_and_group_given(self):
        sdk = SDK(
            project_id=self.project_id,
            group_id=self.group_id,
            username=self.username,
            password=self.password,
        )
        assert sdk._client.project_id == self.project_id

    def test_client_group_and_project_arguments_not_given(self):
        with pytest.raises(
            TypeError,
            match=r"Either a 'group_id' or 'project_id' has to be given as argument",
        ):
            SDK(username=self.username, password=self.password)



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime, timedelta, timezone
from typing import Any, Optional

from auth0.v3.authentication import GetToken  # type: ignore[import-untyped]
from jwt import decode, DecodeError
from requests import PreparedRequest
from requests.auth import AuthBase
from requests.exceptions import HTTPError

from pasqal_cloud.endpoints import Auth0Conf


class HTTPBearerAuthenticator(AuthBase):
    def __init__(self, token_provider: Optional[TokenProvider]):
        if not token_provider:
            raise Exception("The authenticator needs a token provider.")
        self.token_provider = token_provider

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        r.headers["Authorization"] = f"Bearer {self.token_provider.get_token()}"
        return r


class TokenProviderError(Exception):
    pass


class TokenProvider(ABC):
    def __init__(self, *args: list[Any], **kwargs: dict): ...

    @abstractmethod
    def get_token(self) -> str:
        raise NotImplementedError


class ExpiringTokenProvider(TokenProvider, ABC):
    """TokenProvider class which caches a token up to its expiry date.

    This is an abstract class. Create your custom TokenProvider by inheriting from
    this class and defining your own `_query_token` method. See Auth0TokenProvider for
    an implementation example.
    """

    __token_cache: Optional[tuple[datetime, str]] = None
    expiry_window: timedelta = timedelta(minutes=1.0)

    @abstractmethod
    def _query_token(self) -> dict[str, Any]:
        """Return a dictionnary mapping the key 'access_token' to the token value.

        Optionally, your dictionnary may include an 'expires_in' key mapped to the
        duration before expiration of your token (in seconds).
        """
        raise NotImplementedError

    def get_token(self) -> str:
        """Returns the token stored in the cache. If token has expired, first refresh
        the cache by calling the _query_token method.
        """
        if self.__token_cache:
            expiry, token = self.__token_cache
            if expiry - self.expiry_window > datetime.now(tz=timezone.utc):
                return token

        self.__token_cache = self._refresh_token_cache()
        return self.__token_cache[1]

    def _refresh_token_cache(self) -> tuple[datetime, str]:
        try:
            token_response = self._query_token()
            expiry = self._extract_expiry(token_response)

            if expiry is None:
                expiry = datetime.now(tz=timezone.utc)

            return (expiry, token_response["access_token"])
        except HTTPError as err:
            raise TokenProviderError(err)

    @staticmethod
    def _extract_expiry(token_response: dict[str, Any]) -> datetime | None:
        """Extracts the expiry datetime from the token response.

        Assumes that the actual token provided is correct, but might be in an
        unexpected format

        Args:
            token_response: A valid token response dictionary.

        Returns:
            The time the token will expire, or None if it can't be calculated
        """
        expires_in = token_response.get("expires_in")
        if expires_in:
            return datetime.now(tz=timezone.utc) + timedelta(seconds=float(expires_in))

        try:
            # We assume the token is valid, but might not be in an expected format
            token = token_response.get("access_token")

            if isinstance(token, str) and "." in token:
                token_timestamp = decode(
                    token,
                    algorithms=["HS256"],
                    options={"verify_signature": False},
                ).get("exp")
                if token_timestamp:
                    return datetime.fromtimestamp(token_timestamp, tz=timezone.utc)

        except DecodeError:
            # The token is not in a format that could be parsed as a JWT
            pass
        return None


class Auth0TokenProvider(ExpiringTokenProvider):
    """ExpiringTokenProvider implementation which fetches a token from auth0."""

    def __init__(self, username: str, password: str, auth0: Auth0Conf):
        """Initializes the token provider with user credentials and
        an auth0 configuration object

        Args:
            username: name of the user to log in as
            password: password of the user to log in as
            auth0: auth0 configuration object to target the proper auth0 tenant and app
        """
        self.username = username
        self.password = password
        self.auth0 = auth0

        # Makes a call in order to check the credentials at creation
        self.get_token()

    def _query_token(self) -> dict[str, Any]:
        token = GetToken(self.auth0.domain)

        # No client secret required for this Application since
        # "Token Endpoint Authentication Method" set to None
        validated_token: dict[str, Any] = token.login(
            client_id=self.auth0.public_client_id,
            client_secret="",
            username=self.username,
            password=self.password,
            audience=self.auth0.audience,
            scope="openid profile email",
            realm=self.auth0.realm,
            grant_type="http://auth0.com/oauth/grant-type/password-realm",
        )
        return validated_token



from typing import Any, Dict, List, Optional, TypedDict, Union

from pydantic import BaseModel, ConfigDict, PrivateAttr
from requests import HTTPError

from pasqal_cloud.client import Client
from pasqal_cloud.errors import JobCancellingError
from pasqal_cloud.utils.jsend import JobResult


class Job(BaseModel):
    """Class for job data.

    Attributes:
        - runs: Number of times the job should be run.
        - batch_id: ID of the batch which the job belongs to.
        - id: Unique identifier for the job.
        - project_id: ID of the project which the users scheduling the job belong to.
        - status: Status of the job. Possible values are:
            PENDING, RUNNING, DONE, CANCELED, TIMED_OUT, ERROR, PAUSED.
        - _client: A Client instance to connect to PCS.
        - created_at: Timestamp of the creation of the job.
        - updated_at: Timestamp of the last update of the job.
        - errors: Error messages that occurred while processing job.
        - start_timestamp: The timestamp of when the job began processing.
        - end_timestamp: The timestamp of when the job finished processing.
        - full_result: Dictionnary of all the results obtained after complete execution
            of the job. It maps the type of results (e.g. "counter", "raw")
            to the associated execution result.
        - result: Bitstring counter result. Should be equal to `full_results["counter"]`
        - variables: Dictionary of variables of the job.
            None if the associated batch is non-parametrized.
        - group_id (deprecated): Use project_id instead.
    """

    runs: int
    batch_id: str
    id: str
    project_id: str
    status: str
    _client: Client = PrivateAttr(default=None)
    created_at: str
    updated_at: str
    errors: Optional[List[str]] = None
    start_timestamp: Optional[str] = None
    end_timestamp: Optional[str] = None
    _full_result: Optional[JobResult] = PrivateAttr(default=None)
    variables: Optional[Dict[str, Any]] = None
    # Ticket (#622)
    group_id: Optional[str] = None
    parent_id: Optional[str] = None

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    def __init__(self, **data: Any) -> None:
        """
        Workaround to make the private attribute '_client' working
        like we need with Pydantic V2, more information on:
        https://docs.pydantic.dev/latest/concepts/models/#private-model-attributes
        """
        super().__init__(**data)
        self._client = data["_client"]

    @property
    def full_result(self) -> Optional[JobResult]:
        if self._full_result is None:
            self._full_result = self._client.get_job_results(self.id)
        return self._full_result

    @property
    def result(self) -> Optional[Dict[str, Any]]:
        counter_result = None
        if self.full_result:
            counter_result = self.full_result["counter"]
        return counter_result

    def cancel(self) -> Dict[str, Any]:
        """Cancel the current job on the PCS."""
        try:
            job_rsp = self._client.cancel_job(self.id)
        except HTTPError as e:
            raise JobCancellingError(e) from e
        self.status = job_rsp.get("status", "CANCELED")
        return job_rsp


class CreateJob(TypedDict, total=False):
    runs: int
    variables: Union[Dict[str, Any], None]



from __future__ import annotations

from typing import Any, Dict, List, Optional

import requests
from pydantic import BaseModel, ConfigDict, field_validator, PrivateAttr
from pydantic_core.core_schema import ValidationInfo
from requests import HTTPError

from pasqal_cloud.client import Client
from pasqal_cloud.errors import (
    InvalidWorkloadResultsFormatError,
    WorkloadCancellingError,
    WorkloadResultsConnectionError,
    WorkloadResultsDecodeError,
    WorkloadResultsDownloadError,
)


class Workload(BaseModel):
    """Class for workload data.

    Attributes:
        - id: Unique identifier for the workload.
        - project_id: ID of the project which the users scheduling the workload
         belong to.
        - status: Status of the workload. Possible values are:
            PENDING, RUNNING, DONE, CANCELED, TIMED_OUT, ERROR, PAUSED.
        - _client: A Client instance to connect to PCS.
        - backend: The backend used for the workload.
        - workload_type: The type of the workload.
        - config: The config containing all the necessary information for
         the workload to run.
        - created_at: Timestamp of the creation of the workload.
        - updated_at: Timestamp of the last update of the workload.
        - errors: Error messages that occurred while processing workload.
        - start_timestamp: The timestamp of when the workload began processing.
        - end_timestamp: The timestamp of when the workload finished processing.
        - result: Result of the workload.
    """

    id: str
    project_id: str
    status: str
    _client: Client = PrivateAttr(default=None)
    backend: str
    workload_type: str
    config: Dict[str, Any]
    created_at: str
    updated_at: str
    errors: Optional[List[str]] = None
    start_timestamp: Optional[str] = None
    end_timestamp: Optional[str] = None
    result_link: Optional[str] = None
    result: Optional[Dict[str, Any]] = None

    model_config = ConfigDict(
        extra="allow", arbitrary_types_allowed=True, validate_default=True
    )

    def __init__(self, **data: Any) -> None:
        """
        Workaround to make the private attribute '_client' working
        like we need with Pydantic V2, more information on:
        https://docs.pydantic.dev/latest/concepts/models/#private-model-attributes
        """
        super().__init__(**data)
        self._client = data["_client"]

    def cancel(self) -> Dict[str, Any]:
        """Cancel the current job on the PCS."""
        try:
            workload_rsp = self._client.cancel_workload(self.id)
        except HTTPError as e:
            raise WorkloadCancellingError(e) from e
        self.status = workload_rsp.get("status", "CANCELED")
        return workload_rsp

    @field_validator("result")
    def result_link_to_result(
        cls, result: Optional[Dict[str, Any]], info: ValidationInfo
    ) -> Optional[Dict[str, Any]]:
        result_link: Optional[str] = info.data.get("result_link")
        if result or not result_link:
            return result
        try:
            res = requests.get(result_link)
        except HTTPError as e:
            raise WorkloadResultsDownloadError(e) from e
        except requests.ConnectionError as e:
            raise WorkloadResultsConnectionError(e) from e
        try:
            data = res.json()
        except Exception as e:
            raise WorkloadResultsDecodeError from e
        if not isinstance(data, dict):
            raise InvalidWorkloadResultsFormatError(type(data))
        return data



from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field, field_serializer, field_validator

from pasqal_cloud.utils.constants import JobStatus


class BaseJobFilters(BaseModel):
    """
    Base class used by Job related filters for shared fields.
    """

    id: Optional[Union[List[Union[UUID, str]], Union[UUID, str]]] = None
    min_runs: Optional[int] = None
    max_runs: Optional[int] = None
    start_date: Optional[datetime] = None
    end_date: Optional[datetime] = None

    @staticmethod
    def convert_to_list(value: Any) -> Any:
        if not value:
            return None
        if isinstance(value, list):
            return value
        return [value]

    @staticmethod
    def convert_str_to_uuid(value: Any) -> Any:
        if isinstance(value, str):
            try:
                return UUID(value)
            except ValueError:
                raise ValueError(f"Cannot convert {value} to UUID.")
        return value

    @field_serializer("start_date", "end_date", check_fields=False)
    def serialize_datetime_to_isoformat(self, date: datetime) -> str:
        return date.isoformat()


class RebatchFilters(BaseJobFilters):
    """
    Class to provide filters for querying jobs to re-create.

    Setting a value for any attribute of this class will add a filter to the query.

    When using several filters at the same time, the API will return elements who pass
    all filters at the same time:
        - If the filter value is a single element, the API will return jobs whose
          attribute matches the provided value.
        - If the filter value is a list, the API will return jobs whose value for
          this attribute is contained in that list.

    Attributes:
        id: Filter by job IDs
        status: Filter by job statuses,
        min_runs: Minimum number of runs.
        max_runs: Maximum number of runs.
        start_date: Retry jobs created at or after this datetime.
        end_date: Retry jobs created at or before this datetime.
    """

    status: Optional[Union[List[JobStatus], JobStatus]] = None

    @field_validator("id", "status", mode="before")
    @classmethod
    def single_item_to_list_validator(cls, values: Dict[str, Any]) -> Any:
        return cls.convert_to_list(values)

    @field_validator("id", mode="after")
    @classmethod
    def str_to_uuid_validator(cls, values: Dict[str, Any]) -> Dict[str, UUID]:
        for item in values:
            cls.convert_str_to_uuid(item)
        return values

    @field_serializer("status", check_fields=False)
    def status_enum_to_string(self, job_statuses: List[JobStatus]) -> List[str]:
        return [job_status.value for job_status in job_statuses]


class CancelJobFilters(BaseJobFilters):
    """
    Class to provide filters for cancelling a group of jobs.

    Setting a value for any attribute of this class will add a filter to the query.

    When using several filters at the same time, the API will return elements who pass
    all filters at the same time:
        - If the filter value is a single element, the API will return jobs whose
          attribute matches the provided value.
        - If the filter value is a list, the API will return jobs whose value for
          this attribute is contained in that list.

    Attributes:
        id: Filter by job IDs
        min_runs: Minimum number of runs.
        max_runs: Maximum number of runs.
        start_date: Retry jobs created at or after this datetime.
        end_date: Retry jobs created at or before this datetime.
    """

    @field_validator("id", mode="before")
    @classmethod
    def single_item_to_list_validator(cls, values: Dict[str, Any]) -> Any:
        return cls.convert_to_list(values)

    @field_validator("id", mode="after")
    @classmethod
    def str_to_uuid_validator(cls, values: Dict[str, Any]) -> Dict[str, UUID]:
        for item in values:
            cls.convert_str_to_uuid(item)
        return values


class JobFilters(BaseJobFilters):
    """
    Class to provide filters for querying jobs.

    Setting a value for any attribute of this class will add a filter to the query.

    When using several filters at the same time, the API will return elements who pass
    all filters at the same time:
        - If the filter value is a single element, the API will return jobs whose
          attribute matches the provided value.
        - If the filter value is a list, the API will return jobs whose value for
          this attribute is contained in that list.

    Attributes:
        id: Filter by job IDs.
        project_id: Filter by project IDs.
        user_id: Filter by user IDs.
        batch_id: Filter by batch IDs.
        status: Filter by job statuses.
        min_runs: Minimum number of runs.
        max_runs: Maximum number of runs.
        errors: Filter by presence of errors.
        start_date: Retry jobs created at or after this datetime.
        end_date: Retry jobs created at or before this datetime.
    """

    project_id: Optional[Union[List[Union[UUID, str]], Union[UUID, str]]] = None
    user_id: Optional[Union[List[str], str]] = None
    batch_id: Optional[Union[List[Union[UUID, str]], Union[UUID, str]]] = None
    status: Optional[Union[List[JobStatus], JobStatus]] = None
    errors: Optional[bool] = None

    @field_validator("id", "project_id", "user_id", "batch_id", "status", mode="before")
    @classmethod
    def single_item_to_list_validator(cls, values: Dict[str, Any]) -> Any:
        return cls.convert_to_list(values)

    @field_validator("id", "project_id", "batch_id", mode="after")
    @classmethod
    def str_to_uuid_validator(cls, values: Dict[str, Any]) -> Dict[str, UUID]:
        for item in values:
            cls.convert_str_to_uuid(item)
        return values

    @field_serializer("status", check_fields=False)
    def status_enum_to_string(self, job_statuses: List[JobStatus]) -> List[str]:
        return [job_status.value for job_status in job_statuses]


class PaginationParams(BaseModel):
    """
    Class providing parameters for paginating query results.

    Attributes:
        offset: The starting index of the items to return (must be 0 or greater).
        limit: The maximum number of items to return (must be greater than 0 and
               less than or equal to 100).
    """

    offset: int = Field(default=0, strict=True, ge=0)
    limit: int = Field(default=100, strict=True, gt=0, le=100)



from __future__ import annotations

from typing import Any

from auth0.v3.exceptions import Auth0Error

from pasqal_cloud.authentication import Auth0TokenProvider


class FakeAuth0AuthenticationSuccess(Auth0TokenProvider):
    def _query_token(self) -> dict[str, Any]:
        return {
            "access_token": "some_token",
            "id_token": "id_token",
            "scope": "openid profile email",
            "expires_in": 86400,
            "token_type": "Bearer",
        }


class FakeAuth0AuthenticationFailure(Auth0TokenProvider):
    def _query_token(self) -> dict[str, Any]:
        raise Auth0Error(
            status_code=403, error_code=403, message="Wrong email/password"
        )



import contextlib
from typing import Any, Generator
from unittest.mock import patch
from uuid import uuid4

import pytest
import requests
from auth0.v3.exceptions import Auth0Error

from pasqal_cloud import (
    AUTH0_CONFIG,
    Auth0Conf,
    Endpoints,
    PASQAL_ENDPOINTS,
    SDK,
)
from pasqal_cloud._version import __version__ as sdk_version
from pasqal_cloud.authentication import TokenProvider
from tests.test_doubles.authentication import (
    FakeAuth0AuthenticationFailure,
    FakeAuth0AuthenticationSuccess,
)


class TestSDKCommonAttributes:
    project_id = "random_project_id"
    username = "random_username"
    password = "random_password"
    new_core_endpoint = "random_endpoint"
    no_username = ""
    no_password = ""


@patch("pasqal_cloud.client.Auth0TokenProvider", FakeAuth0AuthenticationSuccess)
class TestAuthSuccess(TestSDKCommonAttributes):
    @patch("pasqal_cloud.client.getpass")
    def test_module_getpass_success(self, getpass):
        getpass.return_value = self.password
        SDK(project_id=self.project_id, username=self.username)
        getpass.assert_called_once()

    def test_authentication_success(self):
        SDK(project_id=self.project_id, username=self.username, password=self.password)

    def test_good_token_provider(self):
        SDK(
            project_id=self.project_id,
            token_provider=FakeAuth0AuthenticationSuccess("username", "password", None),
        )

    def test_custom_token_provider(self):
        """Test that the custom provider suggested in the readme is working"""

        class CustomTokenProvider(TokenProvider):
            def get_token(self):
                return "your-token"  # Replace this value with your token

        SDK(token_provider=CustomTokenProvider(), project_id="project_id")

    @pytest.mark.filterwarnings(
        "ignore:The parameters 'endpoints' and 'auth0' are deprecated, from now use"
        " 'env' instead"
    )
    def test_correct_endpoints(self):
        sdk = SDK(
            project_id=self.project_id,
            username=self.username,
            password=self.password,
            endpoints=Endpoints(core=self.new_core_endpoint),
        )
        assert sdk._client.endpoints.core == self.new_core_endpoint

    @pytest.mark.filterwarnings(
        "ignore:The parameters 'endpoints' and 'auth0' are deprecated, from now use"
        " 'env' instead"
    )
    def test_correct_new_auth0(self):
        new_auth0 = Auth0Conf(domain="new_domain")
        SDK(
            project_id=self.project_id,
            username=self.username,
            password=self.password,
            auth0=new_auth0,
        )


@patch("pasqal_cloud.client.Auth0TokenProvider", FakeAuth0AuthenticationFailure)
class TestAuthFailure(TestSDKCommonAttributes):
    @patch("pasqal_cloud.client.getpass")
    def test_module_getpass_bad_password(self, getpass):
        getpass.return_value = self.password

        with pytest.raises(Auth0Error):
            SDK(project_id=self.project_id, username=self.username)

        getpass.assert_called_once()

    def test_module_bad_password(self):
        with pytest.raises(Auth0Error):
            SDK(
                project_id=self.project_id,
                username=self.username,
                password=self.password,
            )


class TestAuthInvalidClient(TestSDKCommonAttributes):
    def test_module_no_user_with_password(self):
        with pytest.raises(
            ValueError,
            match="At least a username or TokenProvider object should be provided",
        ):
            SDK(
                project_id=self.project_id,
                username=self.no_username,
                password=self.password,
            )

    @patch("pasqal_cloud.client.getpass")
    def test_module_no_password(self, getpass):
        getpass.return_value = ""
        with pytest.raises(
            ValueError, match="The prompted password should not be empty"
        ):
            SDK(
                project_id=self.project_id,
                username=self.username,
                password=self.no_password,
            )

    @patch("pasqal_cloud.client.getpass")
    def test_module_getpass_no_password(self, getpass):
        getpass.return_value = self.no_password

        with pytest.raises(
            ValueError, match="The prompted password should not be empty"
        ):
            SDK(project_id=self.project_id, username=self.username)

        getpass.assert_called_once()

    def test_bad_token_provider(self):
        with pytest.raises(TypeError):
            SDK(project_id=self.project_id, token_provider="token")

    def test_bad_auth0(self):
        with pytest.raises(TypeError):
            SDK(
                project_id=self.project_id,
                username=self.username,
                password=self.password,
                auth0="",
            )

    def test_authentication_no_credentials_provided(self):
        with pytest.raises(
            ValueError,
            match="At least a username or TokenProvider object should be provided",
        ):
            SDK(project_id=self.project_id)

    @pytest.mark.filterwarnings(
        "ignore:The parameters 'endpoints' and 'auth0' are deprecated, from now use"
        " 'env' instead"
    )
    def test_bad_endpoints(self):
        with pytest.raises(TypeError):
            SDK(
                project_id=self.project_id,
                username=self.username,
                password=self.password,
                endpoints={
                    "core": "",
                    "account": "",
                },
            )


@patch("pasqal_cloud.client.Auth0TokenProvider", FakeAuth0AuthenticationSuccess)
class TestEnvSDK(TestSDKCommonAttributes):
    @pytest.mark.parametrize(
        ("env", "core_endpoint_expected"),
        [
            ("prod", "https://apis.pasqal.cloud/core-fast"),
            ("preprod", "https://apis.preprod.pasqal.cloud/core-fast"),
            ("dev", "https://apis.dev.pasqal.cloud/core-fast"),
        ],
    )
    def test_select_env(self, env: str, core_endpoint_expected: str):
        sdk = SDK(
            project_id=self.project_id,
            username=self.username,
            password=self.password,
            auth0=AUTH0_CONFIG[env],
            endpoints=PASQAL_ENDPOINTS[env],
        )
        assert sdk._client.endpoints.core == core_endpoint_expected


class TestSDKRetry:
    """
    When we make HTTP calls, certain status codes will either force
    the SDK to retry a HTTP call, return a payload or raise an immediate
    exception.
    """

    @pytest.fixture(autouse=True)
    @patch(
        "pasqal_cloud.client.Auth0TokenProvider",
        FakeAuth0AuthenticationSuccess,
    )
    def _init_sdk(self):
        self.sdk = SDK(
            username="me@test.com",
            password="password",
            project_id=str(uuid4()),
        )

    @pytest.fixture(autouse=True)
    def _mock_sleep(self):
        """
        This fixture overrides sleeps, so tests don't need to wait for
        the entire duration of a sleep command.
        """
        with patch("time.sleep"):
            yield

    def test_download_results_retry_on_exception(
        self, mock_request: Generator[Any, Any, None]
    ):
        """
        Test the retry logic for HTTP calls when an error status code is encountered.

        This test verifies that when an HTTP request fails with an error status code,
        the system retries the HTTP call 2 additional times, resulting in a total of
        3 HTTP calls.
        """

        mock_request.reset_mock()
        mock_request.register_uri(
            "GET", "http://result-link", status_code=500, text="fake-results"
        )
        with contextlib.suppress(Exception):
            self.sdk._client._download_results("http://result-link")
        assert len(mock_request.request_history) == 6

    def test_download_results_retry_on_connection_error(
        self, mock_request: Generator[Any, Any, None]
    ):
        """
        Test the retry logic for HTTP calls when a network error is encountered.

        This test verifies that when an HTTP request fails with
        requests.connectionError, the system retries the HTTP call 2 additional times,
        resulting in a total of 3 HTTP calls.
        """

        def raise_connection_error(*_):
            raise requests.ConnectionError

        mock_request.reset_mock()
        mock_request.register_uri(
            "GET", "http://result-link", body=raise_connection_error
        )

        with contextlib.suppress(requests.ConnectionError):
            self.sdk._client._download_results("http://result-link")
        assert len(mock_request.request_history) == 6

    @pytest.mark.parametrize("status_code", [408, 425, 429, 500, 502, 503, 504])
    def test_sdk_retry_on_exception(
        self, mock_request: Generator[Any, Any, None], status_code: int
    ):
        """
        If a HTTP status code matches any of the codes passed as parameters,
        we should retry a HTTP call 5 more times.

        This test should confirm that 6 HTTP calls take place per "valid retry" status
        code.
        """
        mock_request.reset_mock()
        mock_request.register_uri("GET", "http://test-domain", status_code=status_code)
        with contextlib.suppress(Exception):
            self.sdk._client._authenticated_request("GET", "http://test-domain")
        assert len(mock_request.request_history) == 6

    def test_sdk_doesnt_retry_on_exceptions(
        self, mock_request: Generator[Any, Any, None]
    ):
        """
        If the HTTP status code is not one we consider valid for retires, we should not
        retry any HTTP calls again, since we most likely won't succeeed.

        This test should confirm that if we get a status code we don't want to try
        then our request total should be 1.
        """
        mock_request.reset_mock()
        mock_request.register_uri("GET", "http://test-domain", status_code=400)
        with contextlib.suppress(Exception):
            self.sdk._client._authenticated_request("GET", "http://test-domain")
        assert len(mock_request.request_history) == 1

    def test_sdk_200_avoids_all_exception_handling(
        self, mock_request: Generator[Any, Any, None]
    ):
        """
        We have no need to retry requests on a successful HTTP request, so
        this test confirms that if we receive a 200 success, then we don't try
        more than 1 request.
        """
        mock_request.reset_mock()
        mock_request.register_uri("GET", "http://test-domain", json={}, status_code=200)
        self.sdk._client._authenticated_request("GET", "http://test-domain")
        assert len(mock_request.request_history) == 1


class TestRequestAllPages:
    @pytest.fixture(autouse=True)
    @patch(
        "pasqal_cloud.client.Auth0TokenProvider",
        FakeAuth0AuthenticationSuccess,
    )
    def _init_sdk(self):
        self.sdk = SDK(
            username="me@test.com",
            password="password",
            project_id=str(uuid4()),
        )

    def test_pagination_request_success(self, mock_request: Generator[Any, Any, None]):
        """
        Test requests with pagination with multiple pages.
        This test verifies that the pagination works correctly by simulating
        multiple pages of data.
        It checks that the correct number of requests is made and the
        combined response contains all the expected items.
        """

        mock_request.reset_mock()
        mock_request.register_uri(
            "GET",
            "http://core-test.com/jobs",
            json={
                "data": [{"id": 1}, {"id": 2}],
                "pagination": {"total": 6, "start": 0, "end": 2},
            },
            status_code=200,
        )
        mock_request.register_uri(
            "GET",
            "http://core-test.com/jobs?offset=2",
            json={
                "data": [{"id": 3}, {"id": 4}],
                "pagination": {"total": 6, "start": 2, "end": 4},
            },
            status_code=200,
        )
        mock_request.register_uri(
            "GET",
            "http://core-test.com/jobs?offset=4",
            json={
                "data": [{"id": 5}, {"id": 6}],
                "pagination": {"total": 6, "start": 4, "end": 6},
            },
            status_code=200,
        )

        response = self.sdk._client._request_all_pages(
            "GET", "http://core-test.com/jobs"
        )
        assert len(response) == 6
        assert response == [{"id": number} for number in range(1, 7)]
        assert len(mock_request.request_history) == 3

    def test_pagination_request_changed_total_items_during_query_success(
        self, mock_request: Generator[Any, Any, None]
    ):
        """
        Test request with pagination where the total number of items
        changes during the query.
        This test verifies that the pagination can handle situations where
        the total number of items changes between requests.
        It ensures that the system gracefully handles such changes
        without errors.
        """
        mock_request.reset_mock()
        mock_request.register_uri(
            "GET",
            "http://core-test.com/jobs",
            json={
                "data": [{"id": 1}, {"id": 2}],
                "pagination": {"total": 8, "start": 0, "end": 2},
            },
            status_code=200,
        )
        mock_request.register_uri(
            "GET",
            "http://core-test.com/jobs?offset=2",
            json={
                "data": [{"id": 3}, {"id": 4}],
                "pagination": {"total": 6, "start": 2, "end": 4},
            },
            status_code=200,
        )
        mock_request.register_uri(
            "GET",
            "http://core-test.com/jobs?offset=4",
            json={
                "data": [{"id": 5}],
                "pagination": {"total": 5, "start": 4, "end": 6},
            },
            status_code=200,
        )

        response = self.sdk._client._request_all_pages(
            "GET", "http://core-test.com/jobs"
        )
        assert len(response) == 5
        assert response == [{"id": number} for number in range(1, 6)]
        assert len(mock_request.request_history) == 3

    def test_request_pagination_without_pagination_success(
        self, mock_request: Generator[Any, Any, None]
    ):
        """
        Test request with pagination when there is only a single page of data.
        This test verifies that the pagination request function can handle responses
        without pagination metadata correctly, by returning the data without
        making additional requests.
        """
        mock_request.reset_mock()
        mock_request.register_uri(
            "GET",
            "http://core-test.com/jobs",
            json={
                "data": [{"id": 1}],
            },
            status_code=200,
        )

        response = self.sdk._client._request_all_pages(
            "GET", "http://core-test.com/jobs"
        )
        assert len(response) == 1
        assert response == [{"id": 1}]

class TestHeaders:
    @pytest.fixture(autouse=True)
    @patch(
        "pasqal_cloud.client.Auth0TokenProvider",
        FakeAuth0AuthenticationSuccess,
    )
    def _init_sdk(self):
        self.sdk = SDK(
            username="me@test.com",
            password="password",
            project_id=str(uuid4()),
        )

    def test_user_agent_in_request_headers(
        self, mock_request: Generator[Any, Any, None]
    ):
        """
        Test that the `_authenticated_request` method of
        the client injects the user-agent header properly
        in the headers.
        """
        mock_request.reset_mock()
        mock_request.register_uri(
            "GET",
            "http://core-test.com",
            status_code=200,
            json={
                "ok": True
            }
        )

        _ = self.sdk._client._authenticated_request(
            "GET", "http://core-test.com"
        )
        assert mock_request.last_request.headers["User-Agent"] == (
            f"PasqalCloudSDK/{sdk_version}"
        )



import json
from datetime import datetime
from typing import Any, Dict, Generator, Optional, Union
from unittest.mock import patch
from uuid import UUID, uuid4

import pytest

from pasqal_cloud import Batch, Job, RebatchFilters, SDK
from pasqal_cloud.batch import Batch as BatchModel
from pasqal_cloud.device import BaseConfig, EmuFreeConfig, EmulatorType, EmuTNConfig
from pasqal_cloud.errors import (
    BatchCancellingError,
    BatchClosingError,
    BatchCreationError,
    JobCreationError,
    JobRetryError,
    OnlyCompleteOrOpenCanBeSet,
    RebatchError,
)
from pasqal_cloud.utils.constants import JobStatus
from tests.conftest import mock_core_response
from tests.test_doubles.authentication import FakeAuth0AuthenticationSuccess
from tests.utils import build_query_params


class TestBatch:
    @pytest.fixture()
    def load_mock_batch_json_response(self) -> Dict[str, Any]:
        with open("tests/fixtures/api/v1/batches/_.POST.json") as f:
            return json.load(f)

    @pytest.fixture(autouse=True)
    def _mock_sleep(self):
        """
        This fixture overrides sleeps, so tests don't need to wait for
        the entire duration of a sleep command.
        """
        with patch("time.sleep"):
            yield

    @pytest.fixture(autouse=True)
    @patch(
        "pasqal_cloud.client.Auth0TokenProvider",
        FakeAuth0AuthenticationSuccess,
    )
    def _init_sdk(self):
        self.sdk = SDK(
            username="me@test.com",
            password="password",
            project_id=str(uuid4()),
        )
        self.pulser_sequence = "pulser_test_sequence"
        self.batch_id = "00000000-0000-0000-0000-000000000001"
        self.job_id = "00000000-0000-0000-0000-000000022010"
        self.n_job_runs = 50
        self.job_variables = {
            "Omega_max": 14.4,
            "last_target": "q1",
            "ts": [200, 500],
        }
        self.simple_job_args = {
            "runs": self.n_job_runs,
            "variables": self.job_variables,
        }
        self.job_result = {"1001": 12, "0110": 35, "1111": 1}
        self.job_full_result = {
            "counter": {"1001": 12, "0110": 35, "1111": 1},
            "raw": ["1001", "1001", "0110", "1001", "0110"],
        }

    @pytest.mark.parametrize("emulator", [None] + [e.value for e in EmulatorType])
    def test_create_batch(
        self, emulator: Optional[str], mock_request: Generator[Any, Any, None]
    ):
        """
        When successfully creating a batch, we should be able to assert
        certain fields and values are in the assigned return object.
        """
        batch = self.sdk.create_batch(
            serialized_sequence=self.pulser_sequence,
            jobs=[self.simple_job_args],
            emulator=emulator,
        )
        assert batch.id == self.batch_id
        assert batch.sequence_builder == self.pulser_sequence
        assert not batch.open
        assert batch.complete
        assert batch.ordered_jobs[0].batch_id == batch.id
        assert mock_request.last_request.method == "POST"

    @pytest.mark.parametrize("emulator", [None] + [e.value for e in EmulatorType])
    def test_create_batch_with_complete_raises_warning(
        self, emulator: Optional[str], mock_request: Generator[Any, Any, None]
    ):
        """
        Test that using complete at batch definition is still accepted but will
        trigger a depreceation warning.
        """
        with pytest.warns(DeprecationWarning):
            batch = self.sdk.create_batch(
                serialized_sequence=self.pulser_sequence,
                jobs=[self.simple_job_args],
                emulator=emulator,
                complete=True,
            )
        assert batch.id == self.batch_id
        assert batch.sequence_builder == self.pulser_sequence
        assert not batch.open

    @pytest.mark.parametrize("emulator", [None] + [e.value for e in EmulatorType])
    def test_create_batch_open_and_complete_raises_error(
        self, emulator: Optional[str], mock_request: Generator[Any, Any, None]
    ):
        """
        Test that setting both complete and open values will result in the proper
        error being raised.
        """
        with pytest.raises(OnlyCompleteOrOpenCanBeSet):
            _ = self.sdk.create_batch(
                serialized_sequence=self.pulser_sequence,
                jobs=[self.simple_job_args],
                emulator=emulator,
                complete=True,
                open=True,
            )

    def test_batch_create_exception(
        self, mock_request_exception: Generator[Any, Any, None]
    ):
        """
        Assert the correct exception is raised when failing to create a batch
        and that the correct methods and URLs are used.
        """
        with pytest.raises(BatchCreationError):
            _ = self.sdk.create_batch(
                serialized_sequence=self.pulser_sequence, jobs=[self.simple_job_args]
            )

        assert mock_request_exception.last_request.method == "POST"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/batches"
        )

    @pytest.mark.filterwarnings(
        "ignore:Argument `fetch_results` is deprecated and will be removed "
        "in a future version. Please use argument `wait` instead"
    )
    @pytest.mark.parametrize(("wait", "fetch_results"), [(True, False), (False, True)])
    def test_create_batch_and_wait(
        self, mock_request: Generator[Any, Any, None], wait: bool, fetch_results: bool
    ):
        batch = self.sdk.create_batch(
            serialized_sequence=self.pulser_sequence,
            jobs=[self.simple_job_args],
            wait=wait,
            fetch_results=fetch_results,
        )
        assert (
            batch.id == "00000000-0000-0000-0000-000000000001"
        )  # the batch_id used in the mock data
        assert batch.sequence_builder == self.pulser_sequence
        assert not batch.open
        assert batch.complete
        assert batch.ordered_jobs
        assert batch.ordered_jobs[0].id == self.job_id
        assert batch.ordered_jobs[0].result == self.job_result
        assert batch.ordered_jobs[0].full_result == self.job_full_result
        # Ticket (#704)
        with pytest.warns(
            DeprecationWarning,
            match="'jobs' attribute is deprecated, use 'ordered_jobs' instead",
        ):
            for job_id, job in batch.jobs.items():
                assert self.job_id == job_id
                assert job.result == self.job_result
                assert job.full_result == self.job_full_result
            assert len(batch.jobs) == len(batch.ordered_jobs)
        assert mock_request.last_request.method == "GET"

    @pytest.mark.usefixtures("mock_request")
    def test_get_batch(self, batch: Batch):
        """
        Assert that with a mocked, successful HTTP request our get_batch
        function returns an insance of the Batch model used in the SDK.
        """
        batch_requested = self.sdk.get_batch(batch.id)
        assert isinstance(batch_requested, BatchModel)

    def test_batch_add_jobs(self, mock_request: Generator[Any, Any, None]):
        """
        Test that after successfully creating, and adding jobs, the total number of jobs
        associated with a batch is correct, and the batch id is in the URL that was most
        recently called.
        """
        batch = self.sdk.create_batch(
            serialized_sequence=self.pulser_sequence, jobs=[self.simple_job_args]
        )
        batch.add_jobs([{"runs": self.n_job_runs, "variables": self.job_variables}])
        assert batch.id in mock_request.last_request.url
        assert len(batch.ordered_jobs) == 2

    def test_batch_add_jobs_failure(
        self, batch, mock_request_exception: Generator[Any, Any, None]
    ):
        """
        When trying to add jobs to a batch, we test that we catch
        an exception JobCreationError while later validating the HTTP method
        and URL executed.
        """
        with pytest.raises(JobCreationError):
            _ = batch.add_jobs(
                [{"runs": self.n_job_runs, "variables": self.job_variables}]
            )
        assert mock_request_exception.last_request.method == "POST"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/batches/{batch.id}/jobs"
        )

    def test_batch_add_jobs_and_wait_for_results(
        self,
        batch: Batch,
        mock_request: Generator[Any, Any, None],
        load_mock_batch_json_response: Dict[str, Any],
    ):
        mock_request.reset_mock()
        # Override the batch id so that we load the right API fixtures
        batch.id = "00000000-0000-0000-0000-000000000002"

        # List of job statuses for each request to the get batch endpoint
        # We should keep fetching until jobs are terminated (DONE or ERROR)
        job_statuses = [
            ("PENDING", "PENDING"),  # First call, both jobs are PENDING
            ("RUNNING", "PENDING"),  # Second call, first job has started
            ("DONE", "PENDING"),  # Third call, first job is done
            (
                "DONE",
                "ERROR",
            ),  # Last call, the second job has an ERROR, both jobs are now terminated
        ]

        call_count = 0  # Count calls to the GET endpoint

        def custom_get_batch_mock(request, _):
            nonlocal call_count
            resp = mock_core_response(request)
            # Override with custom status the fixture data
            resp["data"]["jobs"][0]["status"] = job_statuses[call_count][0]
            resp["data"]["jobs"][1]["status"] = job_statuses[call_count][1]
            call_count += 1
            return resp

        mock_request.register_uri(
            "GET", f"/core-fast/api/v1/batches/{batch.id}", json=custom_get_batch_mock
        )

        mock_request.register_uri(
            "POST",
            f"/core-fast/api/v1/batches/{batch.id}/jobs",
            json=load_mock_batch_json_response,
        )
        batch.add_jobs(
            [{"runs": self.n_job_runs}, {"runs": self.n_job_runs}],
            wait=True,
        )

        # Several calls take place to POST /core-fast/api/v1/batches
        # and GET core-fast/api/v1/batches/{id}
        assert mock_request.call_count == 3

        assert mock_request.request_history[0].method == "POST"
        methods = {req.method for req in mock_request.request_history}

        # Check both GET and POST methods execute
        assert {"GET", "POST"} == methods

        assert all(job.result == self.job_result for job in batch.ordered_jobs)
        assert all(
            job.full_result == self.job_full_result for job in batch.ordered_jobs
        )

    @pytest.mark.usefixtures("mock_request")
    def test_batch_declare_complete(self, batch: Batch):
        """
        Test that calling declare_complete on the batch will
        raise a deprecation warning.
        """
        with pytest.warns(DeprecationWarning):
            batch.declare_complete(wait=False)
        assert batch.complete
        assert not batch.open

    @pytest.mark.usefixtures("mock_request")
    def test_batch_close(self, batch: Batch):
        batch.close(wait=False)
        assert batch.complete
        assert not batch.open

    def test_batch_close_failure(
        self, batch: Batch, mock_request_exception: Generator[Any, Any, None]
    ):
        with pytest.raises(BatchClosingError):
            _ = batch.close(wait=False)

        assert batch.status == "PENDING"
        mock_request_exception.stop()

    def test_batch_close_and_wait_for_results(
        self, batch: Batch, mock_request: Generator[Any, Any, None]
    ):
        """TODO"""
        batch.close(wait=True)
        assert batch.complete
        assert not batch.open
        assert mock_request.last_request.method == "GET"
        assert (
            mock_request.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v2/jobs?batch_id={batch.id}"
            "&order_by=ordered_id&order_by_direction=ASC"
        )
        assert batch.ordered_jobs[0].batch_id == batch.id
        assert batch.ordered_jobs[0].result == self.job_result
        assert batch.ordered_jobs[0].full_result == self.job_full_result
        assert len(batch.ordered_jobs) == 1

    def test_cancel_batch_self(self, mock_request, batch):
        batch.cancel()
        assert batch.status == "CANCELED"
        assert mock_request.last_request.method == "PUT"
        assert (
            mock_request.last_request.url == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/batches/{self.batch_id}/cancel"
        )

    def test_cancel_batch_self_error(self, mock_request_exception, batch):
        with pytest.raises(BatchCancellingError):
            batch.cancel()
        assert batch.status == "PENDING"
        assert mock_request_exception.last_request.method == "PUT"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/batches/{self.batch_id}/cancel"
        )

    def test_cancel_batch_sdk(self, mock_request):
        client_rsp = self.sdk.cancel_batch(self.batch_id)
        assert type(client_rsp) == Batch
        assert client_rsp.status == "CANCELED"
        assert mock_request.last_request.method == "PUT"
        assert (
            mock_request.last_request.url == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/batches/{self.batch_id}/cancel"
        )

    def test_cancel_batch_sdk_error(
        self, mock_request_exception: Generator[Any, Any, None]
    ):
        """
        Assert that a BatchCancellingError is raised appropriately for
        failling requests when calling sdk.cancel_batch(...)

        This test also assert the most recently used HTTP method and URL
        are correct.
        """
        with pytest.raises(BatchCancellingError):
            _ = self.sdk.cancel_batch(self.batch_id)

        assert mock_request_exception.last_request.method == "PUT"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/batches/{self.batch_id}/cancel"
        )

    @pytest.mark.parametrize(
        ("emulator", "configuration", "expected"),
        [
            (EmulatorType.EMU_TN, EmuTNConfig(), EmuTNConfig()),
            (None, None, None),
            (
                EmulatorType.EMU_FREE,
                EmuFreeConfig(),
                EmuFreeConfig(extra_config={"dt": 10.0, "precision": "normal"}),
            ),
            (
                "SomethingElse",
                BaseConfig(),
                BaseConfig(extra_config={"dt": 10.0, "precision": "normal"}),
            ),
        ],
    )
    @pytest.mark.usefixtures("mock_request")
    def test_create_batch_configuration(
        self, emulator: str, configuration: BaseConfig, expected: BaseConfig
    ):
        """
        Assert that when creating a batch with a certain confiuration,
        we create the exected object
        """
        batch = self.sdk.create_batch(
            serialized_sequence=self.pulser_sequence,
            jobs=[self.simple_job_args],
            emulator=emulator,
            configuration=configuration,
        )
        assert batch.configuration == expected

    def test_batch_instantiation_with_extra_field(self, batch: Batch):
        """Instantiating a batch with an extra field should not raise an error.

        This enables us to add new fields in the API response on the batches endpoint
        without breaking compatibility for users with old versions of the SDK where
        the field is not present in the Batch class.
        """
        batch_dict = batch.model_dump()  # Batch data expected by the SDK
        # We add an extra field to mimick the API exposing new values to the user
        batch_dict["new_field"] = "any_value"

        new_batch = Batch(**batch_dict)  # this should raise no error
        assert (
            new_batch.new_field == "any_value"
        )  # The new value should be stored regardless

    @pytest.mark.usefixtures("mock_request")
    def test_create_batch_fetch_results_deprecated(
        self,
    ):
        """
        When trying to retrieve batch results with a deprecated argument
        we are able to confirm a certain warning is raised in response.
        """
        with pytest.warns(DeprecationWarning):
            self.sdk.create_batch(
                serialized_sequence=self.pulser_sequence,
                jobs=[self.simple_job_args],
                fetch_results=True,
            )

    @pytest.mark.usefixtures("mock_request")
    def test_get_batch_fetch_results_deprecated(
        self,
    ):
        """Assert that when we use the deprecated argument
        'fetch_results', we receive a DeprecationWarning"""
        with pytest.warns(DeprecationWarning):
            self.sdk.get_batch(self.batch_id, fetch_results=True)

    @pytest.mark.parametrize(
        "filters",
        [
            # No filters provided
            None,
            # Empty object
            RebatchFilters(),
            # Job ID as UUID
            RebatchFilters(id=[UUID(int=0x1)]),
            # Start date
            RebatchFilters(start_date=datetime(2023, 1, 1)),
            # End date
            RebatchFilters(end_date=datetime(2023, 1, 1)),
            # Single job status
            RebatchFilters(status=JobStatus.DONE),
            # List of job statuses
            RebatchFilters(status=[JobStatus.DONE, JobStatus.PENDING]),
            # Minimum runs
            RebatchFilters(min_runs=10),
            # Maximum runs
            RebatchFilters(max_runs=10),
            # Combined
            RebatchFilters(
                id=[UUID(int=0x1)],
                status=[JobStatus.DONE, JobStatus.PENDING],
                min_runs=10,
                max_runs=10,
                start_date=datetime(2023, 1, 1),
                end_date=datetime(2023, 1, 1),
            ),
        ],
    )
    def test_rebatch_success(
        self,
        mock_request: Generator[Any, Any, None],
        filters: Union[RebatchFilters, None],
    ):
        """
        As a user using the SDK with proper credentials,
        I can rebatch an existing batch with specific filters.
        The resulting request will create a new batch which includes
        copies of the jobs that match the filters.
        """
        batch = self.sdk.rebatch(self.batch_id, filters=filters)
        assert mock_request.last_request.method == "POST"

        # Convert filters to the appropriate format for query parameters
        query_params = build_query_params(
            filters.model_dump(exclude_unset=True) if filters is not None else None,
        )

        # Check that the correct url was requested with query params
        assert (
            mock_request.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/batches/"
            + f"{self.batch_id}/rebatch{query_params}"
        )
        assert batch.parent_id == self.batch_id
        assert batch.ordered_jobs[0].parent_id

    def test_rebatch_sdk_error(self, mock_request_exception: Generator[Any, Any, None]):
        """
        As a user using the SDK with proper credentials,
        if my request for rebatching returns a status code
        different from 200,
        I am faced with the RebatchError exception.
        """
        with pytest.raises(RebatchError):
            _ = self.sdk.rebatch(self.batch_id)
        assert mock_request_exception.last_request.method == "POST"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/batches/"
            + f"{self.batch_id}/rebatch"
        )

    def test_retry(
        self,
        mock_request: Generator[Any, Any, None],
    ):
        """
        As a user using the SDK with proper credentials,
        I can retry a job from a given batch.
        The resulting job is added to the batch ordered jobs
        and shares the same variables and results as the original job.
        The endpoint to add jobs was called.
        """
        batch = self.sdk.create_batch(
            serialized_sequence=self.pulser_sequence,
            jobs=[self.simple_job_args],
        )
        assert len(batch.ordered_jobs) == 1
        batch.retry(batch.ordered_jobs[0])
        assert len(batch.ordered_jobs) == 2
        assert batch.ordered_jobs[1].variables == batch.ordered_jobs[0].variables
        assert batch.ordered_jobs[1].runs == batch.ordered_jobs[0].runs
        assert mock_request.last_request.method == "POST"
        assert (
            mock_request.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/batches/"
            + f"{self.batch_id}/jobs"
        )

    def test_retry_sdk_error(
        self,
        batch: Batch,
        job: Job,
        mock_request_exception: Generator[Any, Any, None],
    ):
        """
        As a user using the SDK with proper credentials,
        if my request for retrying a job returns a status code
        different from 200,
        I am faced with the JobRetryError exception.
        """
        batch.ordered_jobs = [job]
        with pytest.raises(JobRetryError):
            batch.retry(job)

        assert len(batch.ordered_jobs) == 1
        assert mock_request_exception.last_request.method == "POST"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/batches/"
            + f"{self.batch_id}/jobs"
        )

    def test_add_jobs_calls_the_correct_features(
        self,
        mock_request: Generator[Any, Any, None],
    ):
        """
        Assert than we calling add_jobs that the correct
        HTTP request method and URL are used.

        We also confirm an instance the correct model
        is returned. The inline mock function allows the reuse
        of an already existing fixture while just overriding a
        the values we want.
        """
        mock_request.reset_mock()

        def inline_mock_batch_response(request, _):
            resp = mock_core_response(request)
            resp["data"]["complete"] = False
            resp["data"]["open"] = True
            resp["data"]["id"] = self.batch_id
            return resp

        # Inline mock assures we reach the end of the function body to call `core`.
        # The add_jobs method has a condition to raise an Exception
        # if the batch is complete.
        mock_request.register_uri(
            "GET",
            f"/core-fast/api/v1/batches/{self.batch_id}",
            json=inline_mock_batch_response,
        )

        b = self.sdk.add_jobs(self.batch_id, [])

        assert (
            mock_request.last_request.url == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/batches/{self.batch_id}/jobs"
        )
        assert isinstance(b, Batch)

    def test_add_jobs_call_raises_job_creation_error(
        self,
        mock_request_exception: Generator[Any, Any, None],
        load_mock_batch_json_response: Generator[Any, Any, None],
    ):
        """
        Assert than we calling add_jobs that the correct
        HTTP request method and URL are used.

        We also confirm that when a batch is marked as closed
        we raise the correct exception. The value
        load_mock_batch_json_response provides a prefabricated data structure
        with a `open: false` value set.
        """
        mock_request_exception.register_uri(
            "POST",
            f"/core-fast/api/v1/batches/{self.batch_id}",
            json=load_mock_batch_json_response,
        )

        with pytest.raises(JobCreationError):
            _ = self.sdk.add_jobs(self.batch_id, [])

        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}"
            f"/api/v1/batches/{self.batch_id}/jobs"
        )

    def test_close_batch_raises_batch_close_error(self):
        """
        Assert that when calling close_batch, we're capable of
        encapsulating the HTTPError as a BatchClosingError instead.
        """
        with pytest.raises(BatchClosingError):
            _ = self.sdk.close_batch(self.batch_id)

    def test_rebatch_raises_value_error_on_invalid_filters(self):
        """
        As a user using the SDK with proper credentials,
        if I pass a dictionary instead of RebatchFilters, a ValueError should be raised.
        """
        with pytest.raises(
            ValueError, match="Filters needs to be a RebatchFilters instance"
        ):
            _ = self.sdk.rebatch(UUID(int=0x1), {"min_runs": 10})



import os

from setuptools import setup

VERSION = "0.1.15"


def get_long_description():
    with open(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md"),
        encoding="utf8",
    ) as fp:
        return fp.read()


setup(
    name="pasqal-sdk",
    description="pasqal-sdk is now pasqal-cloud",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    version=VERSION,
    install_requires=["pasqal-cloud"],
    classifiers=["Development Status :: 7 - Inactive"],
)



from datetime import datetime
from typing import Any, Dict, Generator, List, Union
from unittest.mock import patch
from uuid import UUID, uuid4

import pytest

from pasqal_cloud import (
    CancelJobFilters,
    Job,
    JobCancellingError,
    JobFetchingError,
    JobFilters,
    PaginatedResponse,
    PaginationParams,
    SDK,
)
from pasqal_cloud.utils.constants import JobStatus
from pasqal_cloud.utils.responses import JobCancellationResponse
from tests.test_doubles.authentication import FakeAuth0AuthenticationSuccess
from tests.utils import build_query_params


class TestJob:
    @pytest.fixture(autouse=True)
    @patch(
        "pasqal_cloud.client.Auth0TokenProvider",
        FakeAuth0AuthenticationSuccess,
    )
    def _init_sdk(self):
        self.sdk = SDK(
            username="me@test.com",
            password="password",
            project_id=str(uuid4()),
        )
        self.pulser_sequence = "pulser_test_sequence"
        self.batch_id = "00000000-0000-0000-0000-000000000001"
        self.job_id = "00000000-0000-0000-0000-000000022010"
        self.n_job_runs = 50
        self.job_variables = {
            "Omega_max": 14.4,
            "last_target": "q1",
            "ts": [200, 500],
        }
        self.simple_job_args = {
            "runs": self.n_job_runs,
            "variables": self.job_variables,
        }
        self.job_result = {"1001": 12, "0110": 35, "1111": 1}
        self.job_full_result = {
            "counter": {"1001": 12, "0110": 35, "1111": 1},
            "raw": ["1001", "1001", "0110", "1001", "0110"],
        }

    @pytest.mark.usefixtures("mock_request")
    def test_get_job(self, job: Job):
        """
        Asserting the anticipated object is returned with
        the mocked values when using SDK methods.
        """
        job_requested = self.sdk.get_job(job.id)
        assert job_requested.id == job.id

    def test_get_job_error(
        self, job, mock_request_exception: Generator[Any, Any, None]
    ):
        """
        When attempting to execute get_job, we receive an exception
        which prevents any values being returned.
        We then assert all the correct methods and urls were used.
        """
        with pytest.raises(JobFetchingError):
            _ = self.sdk.get_job(job.id)
        assert mock_request_exception.last_request.method == "GET"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}"
            f"/api/v2/jobs/{job.id}"
        )

    def test_cancel_job_self(self, mock_request: Generator[Any, Any, None], job: Job):
        """
        After cancelling a job, we can assert that the status is CANCELED as expected
        and that the correct methods and URLS were used.
        """
        job.cancel()
        assert job.status == "CANCELED"
        assert mock_request.last_request.method == "PUT"
        assert (
            mock_request.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/jobs/{self.job_id}/cancel"
        )

    def test_cancel_job_self_error(
        self, mock_request_exception: Generator[Any, Any, None], job: Job
    ):
        """
        When trying to cancel a job, we assert that an exception is raised
        while confirming the expected HTTP methods and URLs are used and that
        the job status is still PENDING.
        """
        with pytest.raises(JobCancellingError):
            job.cancel()
        assert job.status == "PENDING"
        assert mock_request_exception.last_request.method == "PUT"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/jobs/{self.job_id}/cancel"
        )

    def test_cancel_job_sdk(self, mock_request: Generator[Any, Any, None]):
        """
        After successfully executing the .cancel_job method,
        we further validate the response object is as anticipated,
        while confirming the expected HTTP methods and URLs are used.
        """
        client_rsp = self.sdk.cancel_job(self.job_id)
        assert type(client_rsp) == Job
        assert client_rsp.status == "CANCELED"
        assert mock_request.last_request.method == "PUT"
        assert (
            mock_request.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/jobs/{self.job_id}/cancel"
        )

    def test_cancel_job_sdk_error(
        self, mock_request_exception: Generator[Any, Any, None]
    ):
        """
        When trying to cancel a job, we assert that an exception is raised
        while confirming the expected HTTP methods and URLs are used.
        """
        with pytest.raises(JobCancellingError):
            _ = self.sdk.cancel_job(self.job_id)
        assert mock_request_exception.last_request.method == "PUT"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/jobs/{self.job_id}/cancel"
        )

    def test_job_instantiation_with_extra_field(self, job):
        """Instantiating a job with an extra field should not raise an error.

        This enables us to add new fields in the API response on the jobs endpoint
        without breaking compatibility for users with old versions of the SDK where
        the field is not present in the Job class.
        """
        job_dict = job.model_dump()  # job data expected by the SDK
        # We add an extra field to mimick the API exposing new values to the user
        job_dict["new_field"] = "any_value"

        new_job = Job(**job_dict)  # this should raise no error
        assert (
            new_job.new_field == "any_value"
        )  # The new value should be stored regardless

    @pytest.mark.parametrize(
        "filters",
        [
            # No filters provided
            None,
            # Empty object
            JobFilters(),
            # Single UUID for id
            JobFilters(id=UUID(int=0x1)),
            # List of UUIDs for id
            JobFilters(id=[UUID(int=0x1), UUID(int=0x2)]),
            # Single string UUID for id
            JobFilters(id=str(UUID(int=0x1))),
            # List of string UUIDs for id
            JobFilters(id=[str(UUID(int=0x1)), str(UUID(int=0x2))]),
            # Single UUID for project_id
            JobFilters(project_id=UUID(int=0x1)),
            # List of UUIDs for project_id
            JobFilters(project_id=[UUID(int=0x1), UUID(int=0x2)]),
            # Single string UUID for project_id
            JobFilters(project_id=str(UUID(int=0x1))),
            # List of string UUIDs for project_id
            JobFilters(project_id=[str(UUID(int=0x1)), str(UUID(int=0x2))]),
            # Single user_id as string
            JobFilters(user_id="1"),
            # List of user_ids as strings
            JobFilters(user_id=["1", "2"]),
            # Single UUID for batch_id
            JobFilters(batch_id=UUID(int=0x2)),
            # List of UUIDs for batch_id
            JobFilters(batch_id=[UUID(int=0x1), UUID(int=0x2)]),
            # Single string UUID for batch_id
            JobFilters(batch_id=str(UUID(int=0x1))),
            # List of string UUIDs for batch_id
            JobFilters(batch_id=[str(UUID(int=0x1)), str(UUID(int=0x2))]),
            # Single status
            JobFilters(status=JobStatus.DONE),
            # List of statuses
            JobFilters(status=[JobStatus.DONE, JobStatus.PENDING]),
            # Minimum runs
            JobFilters(min_runs=10),
            # Maximum runs
            JobFilters(max_runs=20),
            # Errors flag
            JobFilters(errors=True),
            # Start date
            JobFilters(start_date=datetime(2023, 1, 1)),
            # End date
            JobFilters(end_date=datetime(2023, 1, 1)),
            # Combined
            JobFilters(
                id=[UUID(int=0x1), str(UUID(int=0x2))],
                project_id=[UUID(int=0x1), UUID(int=0x2)],
                user_id=["1", "2"],
                batch_id=[UUID(int=0x1), UUID(int=0x2)],
                status=JobStatus.DONE,
                min_runs=10,
                max_runs=20,
                errors=True,
                start_date=datetime(2023, 1, 1),
                end_date=datetime(2023, 1, 1),
            ),
        ],
    )
    def test_get_jobs_with_filters_success(
        self,
        mock_request: Any,
        filters: Union[JobFilters, None],
    ):
        """
        As a user using the SDK with proper credentials,
        I can get a list of jobs with specific filters.
        The resulting request will retrieve the jobs that match the filters.
        """
        response = self.sdk.get_jobs(filters=filters)
        assert isinstance(response, PaginatedResponse)
        assert isinstance(response.total, int)
        assert isinstance(response.results, List)
        for item in response.results:
            assert isinstance(item, Job)
        assert mock_request.last_request.method == "GET"

        # Convert filters to the appropriate format for query parameters
        query_params = build_query_params(
            filters.model_dump(exclude_unset=True) if filters is not None else None,
            PaginationParams().model_dump(),
        )
        # Check that the correct url was requested with query params
        assert (
            mock_request.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v2/jobs{query_params}"
        )

    @pytest.mark.parametrize(
        "pagination",
        [
            None,
            PaginationParams(),
            PaginationParams(limit=1),
            PaginationParams(offset=50, limit=1),
            PaginationParams(offset=100, limit=100),
            PaginationParams(offset=1000, limit=100),
        ],
    )
    def test_get_jobs_with_pagination_success(
        self, mock_request: Any, pagination: PaginationParams
    ):
        """
        As a user using the SDK with proper credentials,
        I can get a list of jobs with pagination parameters.
        The resulting request will retrieve the jobs that match
        the pagination parameters.
        """
        response = self.sdk.get_jobs(pagination_params=pagination)
        assert isinstance(response, PaginatedResponse)
        assert isinstance(response.total, int)
        assert isinstance(response.results, List)
        for item in response.results:
            assert isinstance(item, Job)
        assert mock_request.last_request.method == "GET"

        # Convert filters to the appropriate format for query parameters
        query_params = build_query_params(
            pagination_params=(
                pagination.model_dump()
                if pagination
                else PaginationParams().model_dump()
            )
        )
        # Check that the correct url was requested with query params
        assert (
            mock_request.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v2/jobs{query_params}"
        )

    def test_get_jobs_sdk_error(
        self, mock_request_exception: Generator[Any, Any, None]
    ):
        """
        As a user using the SDK with proper credentials,
        if my request for getting jobs returns a status code different from 200,
        I am faced with the JobFetchingError.
        """
        with pytest.raises(JobFetchingError):
            _ = self.sdk.get_jobs()
        assert mock_request_exception.last_request.method == "GET"
        assert (
            mock_request_exception.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v2/jobs?offset=0&limit=100"
        )

    def test_get_jobs_raises_value_error_on_invalid_filters(self):
        """
        As a user using the SDK with proper credentials,
        if I pass a dictionary instead of JobFilters, a ValueError should be raised.
        """
        with pytest.raises(
            ValueError, match="Filters needs to be a JobFilters instance"
        ):
            _ = self.sdk.get_jobs(filters={"min_runs": 10})

    def test_get_jobs_raises_value_error_on_invalid_pagination_params(self):
        """
        As a user using the SDK with proper credentials,
        if I pass a dictionary instead of PaginationParams, a ValueError should
        be raised.
        """
        with pytest.raises(
            ValueError,
            match="Pagination parameters needs to be a PaginationParams instance",
        ):
            _ = self.sdk.get_jobs(pagination_params={"offset": 100, "limit": 100})

    def test_get_jobs_raises_value_error_on_invalid_value_for_offset(self):
        """
        As a user using the SDK with proper credentials,
        if I pass a 'from_index' inferior to 0, it raises a ValueError to the user.
        """
        with pytest.raises(
            ValueError, match="Input should be greater than or equal to 0"
        ):
            _ = self.sdk.get_jobs(pagination_params=PaginationParams(offset=-1))

    @pytest.mark.parametrize(
        "limit",
        [-1, 0, 101],
    )
    def test_get_jobs_raises_value_error_on_invalid_value_for_limit(self, limit: int):
        """
        As a user using the SDK with proper credentials,
        if I pass a 'limit' inferior to 1 or superior to 100,
        it raises a ValueError to the user.
        """
        with pytest.raises(ValueError, match="limit"):
            _ = self.sdk.get_jobs(pagination_params=PaginationParams(limit=limit))

    @pytest.mark.parametrize(
        "filters",
        [
            # No filters provided
            None,
            # Empty object
            CancelJobFilters(),
            # Single UUID for id
            CancelJobFilters(id=UUID(int=0x1)),
            # List of UUIDs for id
            CancelJobFilters(id=[UUID(int=0x1), UUID(int=0x2)]),
            # Single string UUID for id
            CancelJobFilters(id=str(UUID(int=0x1))),
            # List of string UUIDs for id
            CancelJobFilters(id=[str(UUID(int=0x1)), str(UUID(int=0x2))]),
            # Minimum runs
            CancelJobFilters(min_runs=10),
            # Maximum runs
            CancelJobFilters(max_runs=20),
            # Start date
            CancelJobFilters(start_date=datetime(2023, 1, 1)),
            # End date
            CancelJobFilters(end_date=datetime(2023, 1, 1)),
            # Combined
            CancelJobFilters(
                id=[UUID(int=0x1), str(UUID(int=0x2))],
                min_runs=10,
                max_runs=20,
                start_date=datetime(2023, 1, 1),
                end_date=datetime(2023, 1, 1),
            ),
        ],
    )
    def test_cancel_jobs_success(
        self,
        mock_request: Any,
        filters: Union[CancelJobFilters, None],
    ):
        """
        As a user using the SDK with proper credentials,
        I can cancel of a group of jobs from a batch with specific filters.
        The resulting request will retrieve the jobs that were cancelled and
        the errors for those that could not be cancelled.
        """
        response = self.sdk.cancel_jobs(batch_id=UUID(int=0x1), filters=filters)
        assert isinstance(response, JobCancellationResponse)

        for item in response.jobs:
            assert isinstance(item, Job)

        for k, v in response.errors.items():
            assert isinstance(k, UUID)
            assert isinstance(v, str)

        assert isinstance(response.errors, Dict)

        assert mock_request.last_request.method == "PUT"

        # Convert filters to the appropriate format for query parameters
        query_params = build_query_params(
            filters.model_dump(exclude_unset=True) if filters is not None else None,
        )
        # Check that the correct url was requested with query params
        assert (
            mock_request.last_request.url
            == f"{self.sdk._client.endpoints.core}/api/v1/batches/{UUID(int=0x1)}"
            f"/cancel/jobs{query_params}"
        )

    def test_cancel_jobs_raises_value_error_on_invalid_filters(self):
        """
        As a user using the SDK with proper credentials,
        if I pass a dictionary instead of CancelJobFilters, a ValueError should
        be raised.
        """
        with pytest.raises(
            ValueError, match="Filters needs to be a CancelJobFilters instance"
        ):
            _ = self.sdk.cancel_jobs(batch_id=UUID(int=0x1), filters={"min_runs": 10})



import functools
import time
from typing import Callable, Iterable, Optional, Tuple, Type, TypeVar

from requests import HTTPError
from typing_extensions import ParamSpec

Param = ParamSpec("Param")
RT = TypeVar("RT")  # return type


def retry_http_error(
    max_retries: int = 5,
    retry_status_code: Optional[Iterable[int]] = None,
    retry_exceptions: Optional[Tuple[Type[Exception]]] = None,
) -> Callable[[Callable[..., RT]], Callable[..., RT]]:
    """
    Decorator to retry an HTTP call when an HTTPError is encountered.

    Args:
        max_retries: The maximum number of retry attempts
        retry_status_codes: Specific HTTP status codes to trigger a retry.
            - NB: If None, retries on all HTTP error status codes.
        retry_exceptions:  List of specific Exception classes that trigger a retry.
            - NB: If None, retries will only occur based on status codes, not
            exceptions.
    """

    def decorator(func: Callable[..., RT]) -> Callable[..., RT]:
        @functools.wraps(func)
        def wrapper(*args: Param.args, **kwargs: Param.kwargs) -> RT:
            for iteration in range(max_retries + 1):
                try:
                    response = func(*args, **kwargs)
                except HTTPError as e:  # noqa:  PERF203
                    if (
                        e.response is None
                        or (
                            retry_status_code is not None
                            and e.response.status_code not in retry_status_code
                        )
                        or iteration == max_retries
                    ):
                        raise e
                    delay = (1 * 2) ** iteration
                    time.sleep(delay)
                except Exception as e:
                    if (
                        retry_exceptions
                        and isinstance(e, retry_exceptions)
                        and iteration < max_retries
                    ):
                        delay = (1 * 2) ** iteration
                        time.sleep(delay)
                    else:
                        raise e
                else:
                    return response

            # There is no scenario where we want to reach this
            # so we can raise a generic Exception
            raise Exception(
                "HTTP Client has encountered an issue it is unable to recover from."
            )

        return wrapper

    return decorator



from typing import Any, Dict, List
from uuid import UUID

from pydantic import BaseModel

from pasqal_cloud.job import Job


class PaginatedResponse(BaseModel):
    """
    Class representing a paginated response structure.

    Attributes:
        total: The total number of items matching the query.
        offset: The starting point for the current set of paginated results.
                It indicates the number of items skipped before the current set.
        results: A list of items that match the query and pagination parameters
                 provided.
    """

    total: int
    offset: int
    results: List[Any]


class JobCancellationResponse(BaseModel):
    """
    Class representing a bulk job cancellation response structure.

    Attributes:
        jobs: A list of jobs that were successfully cancelled.
        errors: A dict of jobs id with the errors explaining why they could
        not be cancelled.
    """

    jobs: List[Job]
    errors: Dict[UUID, str]
